@@ -5,12 +5,12 @@
 (defparameter *autoconf-prefix* "/usr/local")
 (defparameter *autoconf-exec_prefix* "/usr/local")
 (defparameter *autoconf-package* "maxima")
-(defparameter *autoconf-version* "5.41.0")
+(defparameter *autoconf-version* "5.42.0")
 (defparameter *autoconf-libdir* "/usr/local/lib")
 (defparameter *autoconf-libexecdir* "/usr/local/libexec")
 (defparameter *autoconf-datadir* "/usr/local/share")
 (defparameter *autoconf-infodir* "/usr/local/info")
-(defparameter *autoconf-host* "powerpc-unknown-openbsd6.3")
+(defparameter *autoconf-host* "arm-unknown-openbsd6.4")
 ;; This variable is kept for backwards compatibiliy reasons:
 ;; We seem to be in the fortunate position that we sometimes need to check for windows.
 ;; But at least until dec 2015 we didn't need to check for a specific windows flavour.
@@ -21,5 +21,5 @@
 ;; This will be T if this was a lisp-only build
 (defparameter *autoconf-lisp-only-build* (eq t '@lisp_only_build@))
  
-(defparameter *maxima-source-root* "/usr/obj/ports/maxima-5.41.0/maxima-5.41.0")
+(defparameter *maxima-source-root* "/usr/obj/ports/maxima-5.42.0/maxima-5.42.0")
 (defparameter *maxima-default-layout-autotools* "true")
